/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau;

import java.applet.Applet;
import java.awt.Dimension;
import java.awt.Font;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.StringReader;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Collection;

import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.text.BadLocationException;

import at.jku.risc.stout.nau.algo.DebugLevel;
import at.jku.risc.stout.nau.algo.Equivariance;
import at.jku.risc.stout.nau.algo.EquivarianceProblem;
import at.jku.risc.stout.nau.algo.EquivarianceSystem;
import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.InputParser;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.util.ControlledException;

/**
 * Simple presentation layer for embedding the rule based algorithm
 * {@linkplain EquivarianceSystem} in the browser.
 * 
 * @author Alexander Baumgartner
 */
public class EquivarianceApplet extends Applet {
	private static final long serialVersionUID = 1L;

	private JTextArea textArea = new JTextArea();
	private JScrollPane scroll = new JScrollPane(textArea,
			JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

	private static final int MAX_LINES = 5000;
	// Jitter for MAX_LINES = 4900 - 5100
	private static final int MAX_LINES_VARIANZ = 100;
	private static final String PRINT_HEAD = "";
	private static final String PRINT_TAIL = "DONE";

	private final PrintStream out = new ConsoleStream("", true);
	private final PrintStream info = new ConsoleStream("INFO:  ", false);
	private final PrintStream err = new ConsoleStream("ERROR: ", false);

	@Override
	public void init() {
		textArea.setEditable(false);
		scroll.setPreferredSize(new Dimension(getWidth(), getHeight() - 5));
		textArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, getHeight() / 20));
		add(scroll);
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			SwingUtilities.updateComponentTreeUI(this);
		} catch (Exception isOk_UseDefaultLookAndFeel) {
		}
	}

	@Override
	public void resize(int width, int height) {
		super.resize(width, height);
		scroll.setPreferredSize(new Dimension(width, height - 5));
		textArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, getHeight() / 20));
	}

	@Override
	public void start() {
		display(PRINT_HEAD);
		try {
			String unifProblem = getParameter("problemset");
			if (unifProblem == null || unifProblem.length() == 0) {
				err.println("No equivariance problem sent by the browser.\n\n"
						+ "This might be a communication-error between your browser and the Java-Applet.\n"
						+ "Please try it once more. If you use Opera, please turn off the Turbo mode.\n"
						+ "If the problem persists, please try another browser.");
			} else {
				boolean justify = "true".equals(getParameter("justify"));
				String nabla = getParameter("nabla");
				if (nabla == null)
					nabla = "";
				DebugLevel debugLevel = DebugLevel.SIMPLE;
				try {
					debugLevel = DebugLevel.valueOf(getParameter("debugmode"));
				} catch (Exception e) {
				}

				final NodeFactory factory = new NodeFactory();
				EquationSystem<EquivarianceProblem> eqSys = new EquationSystem<EquivarianceProblem>() {
					@Override
					public EquivarianceProblem newEquation(NominalTerm l,
							NominalTerm r) {
						return new EquivarianceProblem(l, r);
					}
				};
				InputParser parser = new InputParser(factory);
				parser.parseEquationSystem(new StringReader(unifProblem), eqSys);
				FreshnessCtx nablaIn = parser
						.parseNabla(new StringReader(nabla));
				Collection<? extends Atom> atomSet = factory
						.getAllByType(factory.classAtom);

				info.println("Justify computed permutation = " + justify);
				info.println("Output format = " + debugLevel);
				info.println();
				info.println("ATOMS:");
				info.println(atomSet);
				info.println("NABLA_IN:");
				nablaIn.printString(info);
				info.println();
				info.println("PROBLEM:");
				for (EquivarianceProblem aup : eqSys) {
					aup.printString(info);
					info.println();
				}
				info.println();

				Equivariance equiv = new Equivariance(eqSys, atomSet, nablaIn);
				equiv.compute(factory, justify, debugLevel, out);
			}
		} catch (ControlledException ex) {
			err.println(ex.getMessage() + " (" + ex.getClass().getSimpleName()
					+ ")");
		} catch (Throwable ex) {
			ex.printStackTrace(err);
		}
		display(PRINT_TAIL);
		try {
			Thread.sleep(300);
		} catch (InterruptedException ignored) {
		}
	}

	private void display(String str) {
		try {
			if (textArea.getLineCount() >= MAX_LINES + MAX_LINES_VARIANZ) {
				textArea.replaceRange(null, 0,
						textArea.getLineEndOffset(MAX_LINES_VARIANZ * 2));
			}
			textArea.append(str);
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
	}

	/**
	 * May be called by JavaScript to enable copying of text from the applet to
	 * the system clipboard.
	 */
	public String getSelectedText() {
		String text = textArea.getSelectedText();
		return text == null ? "" : text;
	}

	class ConsoleStream extends PrintStream {
		ConsoleStream(final String prefix, final boolean isCount) {
			super(new OutputStream() {
				private final NumberFormat LINENUM_FORMAT = new DecimalFormat(
						"000000: ");
				long count = 0L;
				StringBuilder buffer = resetBuffer(new StringBuilder());

				@Override
				public void write(int b) throws IOException {
					buffer.appendCodePoint(b);
					if (b == '\n') {
						count++;
						flush();
						resetBuffer(buffer);
					}
				}

				private StringBuilder resetBuffer(StringBuilder buffer) {
					if (isCount)
						buffer.append(LINENUM_FORMAT.format(count));
					buffer.append(prefix);
					return buffer;
				}

				@Override
				public void flush() throws IOException {
					display(buffer.toString());
					buffer.setLength(0);
				}
			});
		}
	}
}
