/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau;

import java.io.PrintStream;
import java.io.StringReader;
import java.util.Arrays;
import java.util.Collection;

import at.jku.risc.stout.nau.algo.DebugLevel;
import at.jku.risc.stout.nau.algo.Equivariance;
import at.jku.risc.stout.nau.algo.EquivarianceProblem;
import at.jku.risc.stout.nau.algo.EquivarianceSystem;
import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.InputParser;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.NominalTerm;

/**
 * Simple presentation layer for running the rule based algorithm
 * {@linkplain EquivarianceSystem} in a shell.
 * 
 * @author Alexander Baumgartner
 */
public class EquivarianceMain {

	static DebugLevel debugLevel = DebugLevel.SIMPLE;
	static boolean justify = false;

	public static void main(String[] args) {
		if (args.length >= 1) {
			if ("-d".equals(args[0])) {
				debugLevel = DebugLevel.PROGRESS;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-v".equals(args[0])) {
				if (debugLevel != DebugLevel.PROGRESS)
					debugLevel = DebugLevel.VERBOSE;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-h".equals(args[0])) {
				printHelp(System.out);
			} else if ("-j".equals(args[0])) {
				justify = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else {
				try {
					PrintStream out = System.out;
					final NodeFactory factory = new NodeFactory();
					EquationSystem<EquivarianceProblem> eqSys = new EquationSystem<EquivarianceProblem>() {
						@Override
						public EquivarianceProblem newEquation(NominalTerm l,
								NominalTerm r) {
							return new EquivarianceProblem(l, r);
						}
					};
					InputParser parser = new InputParser(factory);
					parser.parseEquationSystem(new StringReader(args[0]), eqSys);
					FreshnessCtx nabla = new FreshnessCtx();
					Collection<? extends Atom> atomSet = factory
							.getAllByType(factory.classAtom);
					if (args.length > 1) {
						nabla = parser.parseNabla(new StringReader(args[1]));
						if (args.length > 2)
							atomSet = parser.parseAtomSet(new StringReader(
									args[2]));
					}

					out.println("Justify computed permutation = "
							+ justify);
					out.println("Output format = " + debugLevel);
					out.println();
					out.println("ATOMS:");
					out.println(atomSet);
					out.println("NABLA_IN:");
					nabla.printString(out);
					out.println();
					out.println("PROBLEM:");
					eqSys.printString(out);
					out.println();
					out.println();

					Equivariance equiv = new Equivariance(eqSys, atomSet, nabla);
					equiv.compute(factory, justify, debugLevel, out);

					out.println();
					out.println();
					out.println("Supported by the Austrian Science Fund (FWF) "
							+ "under the project SToUT P 24087-N18");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} else {
			System.out.println("Insufficient parameters!");
			System.out.println();
			printHelp(System.out);
		}
	}

	private static void printHelp(PrintStream out) {
		System.out.println("Equivariance problem solver for nominal terms");
		System.out.println();
		System.out
				.println("Give: A set of equations of nominal terms of the form:");
		System.out.println("      \"s1 = t1; s2 = t2; ...; sn = tn\";");
		System.out.println("      A freshness context of the form:");
		System.out.println("      \"a#X, b#Y, c#Z, ...\";");
		System.out.println("      Some atoms:");
		System.out.println("      \"d,e,f ...\";");
		System.out.println("Input example: \"f(a,a) = f(b,b)\"");
		System.out.println("     Or typed: \"f(a:i,a:i):o = f:i-i-o(b,b)\"");
		System.out.println("Or with nabla and atoms:");
		System.out.println("\"f(a:i,a:i):o> = f:i-i-o(b,b)\" "
				+ "\"a#X, b#X\" \"a, b, c, d\"");
		System.out.println();
		System.out.println("Simple execution example: "
				+ "java at.jku.risc.stout.nau.EquivarianceMain -j "
				+ "\"f(a,b) = f(b,c)\"");
		System.out.println("Advanced execution example with jvm parameters:");
		System.out.println("java -Xmx2048m -XX:MaxPermSize=1024m -Xss128m "
				+ "at.jku.risc.stout.nau.EquivarianceMain -j -d "
				+ "\"f(a,b,a.X) = f(b,c,c.X)\" "
				+ "\"a#X, b#X, c#X\" \"a, b, c, d\"");
		System.out.println();
		System.out.println("  -j    justify computed permutation");
		System.out
				.println("  -v    verbose output mode prints some more information");
		System.out
				.println("  -d    debug mode prints out every available information");
		System.out.println("  -h    print this information");
	}
}
