/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau;

import java.io.StringReader;
import java.util.Collection;

import at.jku.risc.stout.nau.algo.DebugLevel;
import at.jku.risc.stout.nau.algo.Equivariance;
import at.jku.risc.stout.nau.algo.EquivarianceProblem;
import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.InputParser;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.NominalTerm;

/**
 * Some test cases
 * 
 * @author Alexander Baumgartner
 */
public class TestEquivariance {

	public static void main(String[] args) {
		test("f(a,b) = f(a,b)");
		test("f(a.a,b) = f(a.a,a)");
		test("f(a,b) = f(b,a)");
		test("f(a:i,b:i):o = f:i-i-o(a,b)");
		test("f(a:i,a:i):o = f:i-i-o(b,b)");
		test("f(a,a) = f(b,b)");
		test("f(a,a) = f(b,c)");
		test("f(a,b) = f(b,c)");
		test("f(a,b,a.X) =^= f(b,c,c.X)", "a#X, b#X, c#X", "a, b, c, d");
		test("a.a = b.c");
		test("a.b = b.a");
		test("a.b = b.a", "", "a,b,c");
		test("h(X1, c.X1) =^= h(X1, c.X1)");
		test("h(X1, a.(a c)X1) =^= h(X1, b.(b c)X1)", "a#X1,b#X1,c#X1");
		test("h(X1, a.X2) =^= h(X1, a.X2)", "a#X1");
		test("a.f(a,b) =^= b.f(b,a)", "");
		test("a.f(a,b) =^= b.f(b,a)", "", "a,b,c");
		test("a.b.c =^= c.a.b");
		test("a.b.b =^= b.b.a");
		test("f(a.b, X) =^= f(b.a, X)", "c#X");
		test("h(X1, a.a.(a c)X1) =^= h(X1, c.b.(b c)X1)", "{a#X1,b#X1,c#X1}");
	}

	public static void test(String eqSysIn) {
		test(eqSysIn, "");
	}

	public static void test(String eqSysIn, String nablaIn) {
		test(eqSysIn, nablaIn, null);
	}

	public static void test(String eqSysIn, String nablaIn, String atoms) {
		try {
			final NodeFactory factory = new NodeFactory();
			EquationSystem<EquivarianceProblem> eqSys = new EquationSystem<EquivarianceProblem>() {
				@Override
				public EquivarianceProblem newEquation(NominalTerm l,
						NominalTerm r) {
					return new EquivarianceProblem(l, r);
				}
			};
			InputParser parser = new InputParser(factory);
			parser.parseEquationSystem(new StringReader(eqSysIn), eqSys);
			FreshnessCtx nabla = parser.parseNabla(new StringReader(nablaIn));
			Collection<? extends Atom> atomSet;
			if (atoms == null)
				atomSet = factory.getAllByType(factory.classAtom);
			else
				atomSet = parser.parseAtomSet(new StringReader(atoms));
			Equivariance equiv = new Equivariance(eqSys, atomSet, nabla);
			eqSys.printString(System.out);
			System.out.println();
			System.out.print("Nabla in: ");
			nabla.printString(System.out);
			System.out.print(" Atoms: ");
			System.out.println(atomSet);
			equiv.compute(factory, true, DebugLevel.PROGRESS, System.out);
			System.out.println();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
