/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.algo;

import java.io.PrintStream;
import java.util.Map.Entry;

import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.NominalPair;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Variable;
import at.jku.risc.stout.nau.util.ControlledException;

/**
 * This class encapsulates the rule based system {@linkplain AntiUnifySystem}
 * and allows to justify the result by back substitution of the generalization
 * with the reported differences inside the computed store.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnify {
	private DebugLevel debugLevel;
	private EquationSystem<AntiUnifyProblem> eqSysOrig;
	private FreshnessCtx nablaIn;
	private AntiUnifySystem nauSys;

	/**
	 * Creates and initializes the encapsulated rule based system
	 * {@linkplain AntiUnifySystem} with the given equation system and freshness
	 * context.
	 */
	public AntiUnify(EquationSystem<AntiUnifyProblem> eqSys,
			FreshnessCtx nablaIn, DebugLevel debugLevel, NodeFactory factory) {
		this.debugLevel = debugLevel;
		this.eqSysOrig = eqSys;
		this.nablaIn = nablaIn;
		this.nauSys = new AntiUnifySystem(factory, eqSys.deepCopy(), nablaIn);
	}

	/**
	 * Calls the rule based system {@linkplain AntiUnifySystem} and tries to
	 * justify the computed result if the second argument of this method is
	 * true.
	 */
	public void antiUnify(boolean justify, PrintStream out)
			throws ControlledException {
		nauSys.compute(debugLevel, out);
		if (debugLevel == DebugLevel.VERBOSE) {
			out.println("  Store: " + nauSys.getStore());
			out.println("  Nabla: " + nauSys.getNablaGen());
			out.println("  Sigma: " + nauSys.getSigma());
		}
		if (debugLevel != DebugLevel.SILENT) {
			out.println("-----------");
			out.println(" Result: " + nauSys);
			out.println("-----------");
			out.println();
		}
		if (justify) {
			if (debugLevel == DebugLevel.PROGRESS)
				out.println("Start justification.");
			for (AntiUnifyProblem aup : eqSysOrig)
				justifyAup(out, aup);
		}
		for (Entry<Variable, NominalTerm> solved : nauSys.getSigma()
				.getMapping().entrySet())
			callback(nauSys, solved.getKey());
	}

	/**
	 * This callback function will be invoked for every found generalization. By
	 * default it does nothing. For instance, one can use this callback function
	 * to collect all the results or to filter out some interesting results.<br>
	 * In combination with a matching algorithm, one can obtain the minimal
	 * complete set of generalizations by using this callback function.
	 */
	public void callback(AntiUnifySystem result, Variable generalizationVar) {
	}

	private void justifyAup(PrintStream out, AntiUnifyProblem aup)
			throws ControlledException {
		FreshnessCtx nablaOut = nauSys.getNablaGen();
		NominalTerm result = nauSys.getSigma().get(aup.getGeneralizationVar());
		if (result == null)
			throw new JustificationException("No result found for " + aup);
		NominalPair lPairOut = new NominalPair(nablaOut.deepCopy(),
				result.deepCopy());
		NominalPair rPairOut = new NominalPair(nablaOut.deepCopy(),
				result.deepCopy());
		for (int j = nauSys.getStore().size() - 1; j >= 0; j--) {
			AntiUnifyProblem backSubsti = nauSys.getStore().get(j);
			Variable var = backSubsti.getGeneralizationVar();
			lPairOut.substitute(var, backSubsti.getLeft());
			rPairOut.substitute(var, backSubsti.getRight());
		}
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("Generating two terms-in-context from the computed generalization");
			out.println("by generating two substitutions from the store and applying them.");
			out.println("Generated left hand side: " + lPairOut);
			out.println("Generated right hand side: " + rPairOut);
		}
		if (debugLevel == DebugLevel.PROGRESS)
			out.println("Testing generated nabla.");
		if (!lPairOut.getFreshEnv().isSubsetOf(nablaIn)
				|| !rPairOut.getFreshEnv().isSubsetOf(nablaIn)) {
			out.println("ERROR! Nabla out is not a subset of nabla in.");
			throw new JustificationException(
					"The computed result is not a valid generalization");
		}
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("OK. Computed nablas are subsets of nabla in.");
			out.println("Testing alpha equivalence of the generated terms with the input terms.");
		}
		if (!lPairOut.getTerm().equivalent(aup.getLeft(), nablaIn)
				|| !rPairOut.getTerm().equivalent(aup.getRight(), nablaIn)) {
			out.println("ERROR! At least one of the generated terms is not alpha "
					+ "equivalent to the corresponding input term under nabla in.");
			throw new JustificationException(
					"The computed result is not a valid generalization");
		}
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("Ok. " + nablaIn + " |-- " + lPairOut.getTerm() + " = "
					+ aup.getLeft());
			out.println("Ok. " + nablaIn + " |-- " + rPairOut.getTerm() + " = "
					+ aup.getRight());
		}
		if (debugLevel != DebugLevel.SILENT && debugLevel != DebugLevel.SIMPLE) {
			out.println("Justification ok!");
		}
	}

	/**
	 * Returns the original equation system.
	 */
	public EquationSystem<AntiUnifyProblem> getEqSysOrig() {
		return eqSysOrig;
	}

	/**
	 * Returns the entire {@linkplain AntiUnifySystem} which is encapsulated by
	 * this class.
	 */
	public AntiUnifySystem getNauSys() {
		return nauSys;
	}
}
