/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.algo;

import java.io.IOException;
import java.io.Writer;

import at.jku.risc.stout.nau.data.Equation;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Variable;

/**
 * This class represents an anti-unification problem (AUP) which consists of one
 * generalization variable (the most general generalization) and an
 * anti-unification {@linkplain Equation}.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifyProblem extends Equation<NominalTerm> {
	private Variable generalizationVar;
	public static String varSeparator = ": ";
	public static String eqSeparator = " =^= ";

	/**
	 * A constructor who takes two nominal terms and a factory to instantiate a
	 * fresh generalization variable.
	 */
	public AntiUnifyProblem(NominalTerm left, NominalTerm right,
			NodeFactory factory) {
		this(left, right, factory.obtainFreshVar(left.getSort()));
	}

	private AntiUnifyProblem(NominalTerm left, NominalTerm right,
			Variable generalizationVar) {
		super(left, right);
		this.generalizationVar = generalizationVar;
	}

	/**
	 * The most general generalization.
	 */
	public Variable getGeneralizationVar() {
		return generalizationVar;
	}

	@Override
	public Equation<NominalTerm> deepCopy() {
		return new AntiUnifyProblem(getLeft().deepCopy(),
				getRight().deepCopy(), generalizationVar);
	}

	@Override
	public void printString(Writer toPrint) throws IOException {
		printString(toPrint, getLeft(), getRight(), generalizationVar);
	}

	public static void printString(Writer toPrint, NominalTerm left,
			NominalTerm right, Variable genVar) throws IOException {
		toPrint.write(String.valueOf(genVar));
		toPrint.write(varSeparator);
		left.printString(toPrint);
		toPrint.write(eqSeparator);
		right.printString(toPrint);
	}
}