/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.algo;

import java.io.IOException;
import java.io.PrintStream;
import java.io.Writer;
import java.util.Collection;
import java.util.List;

import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.Abstraction;
import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.FunctionApplication;
import at.jku.risc.stout.nau.data.atom.HasSort;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Permutation;
import at.jku.risc.stout.nau.data.atom.Sort;
import at.jku.risc.stout.nau.data.atom.Suspension;
import at.jku.risc.stout.nau.data.atom.Variable;
import at.jku.risc.stout.nau.util.ControlledException;
import at.jku.risc.stout.nau.util.DataStructureFactory;
import at.jku.risc.stout.nau.util.Printable;

/**
 * This class represents a rule based system for Nominal Anti-Unificaion.<br>
 * The algorithm P is described in the paper:<br>
 * <a href="http://www.risc.jku.at/home/abaumgar/">Alexander Baumgartner</a>, <a
 * href="http://www.risc.jku.at/people/tkutsia/">Temur Kutsia</a>, <a
 * href="http://www.iiia.csic.es/~levy/">Jordi Levy</a> and <a
 * href="http://ima.udg.edu/~villaret/">Mateu Villaret</a>, Nominal
 * Anti-Unificaion
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifySystem extends Printable {
	private EquationSystem<AntiUnifyProblem> problemSet;
	private FreshnessCtx nablaIn;
	private Collection<? extends Atom> atoms;
	private List<AntiUnifyProblem> store;
	private FreshnessCtx nablaGen;
	private Substitution sigma;

	private EquivarianceSystem permSys;

	/**
	 * Creates and initializes the rule based system. Uses an empty store, empty
	 * nablaGen and empty sigma.
	 * 
	 * @see AntiUnify
	 */
	public AntiUnifySystem(NodeFactory factory,
			EquationSystem<AntiUnifyProblem> problemSet, FreshnessCtx nablaIn) {
		this(factory, problemSet, nablaIn, factory
				.getAllByType(factory.classAtom), NEW_STORE(),
				new FreshnessCtx(), new Substitution());
	}

	/*
	 * Type inference in java is very silly ==> We need this stupid method.
	 */
	private static List<AntiUnifyProblem> NEW_STORE() {
		return DataStructureFactory.$.newList();
	}

	/**
	 * Creates and initializes the rule based system.
	 * 
	 * @see AntiUnify
	 */
	public AntiUnifySystem(NodeFactory factory,
			EquationSystem<AntiUnifyProblem> problemSet, FreshnessCtx nablaIn,
			Collection<? extends Atom> atoms, List<AntiUnifyProblem> store,
			FreshnessCtx nablaGen, Substitution sigma) {
		this.problemSet = problemSet;
		this.nablaIn = nablaIn;
		this.atoms = atoms;
		this.store = store;
		this.nablaGen = nablaGen;
		this.sigma = sigma;
		this.permSys = new EquivarianceSystem(factory);
		for (AntiUnifyProblem aup : problemSet) {
			// Create X -> X for all generalization variables.
			// The algorithm we will only modify the range of sigma.
			sigma.put(aup.getGeneralizationVar(),
					new Suspension(aup.getGeneralizationVar()));
		}
	}

	/**
	 * Computes the result silently.
	 */
	public void compute() throws ControlledException {
		compute(DebugLevel.SILENT, null);
	}

	/**
	 * Computes the result and prints progress information to
	 * {@linkplain PrintStream} according to {@linkplain DebugLevel}.
	 */
	public void compute(DebugLevel debugLevel, PrintStream debugOut)
			throws ControlledException {
		while (!problemSet.isEmpty()) {
			debug(debugLevel, debugOut);

			AntiUnifyProblem eq = problemSet.popFirst();
			Variable genVar = eq.getGeneralizationVar();
			NominalTerm lTerm = eq.getLeft();
			NominalTerm rTerm = eq.getRight();
			HasSort<? extends Sort> lHead = lTerm.getHead();
			HasSort<? extends Sort> rHead = rTerm.getHead();

			if (lTerm.getClass() == rTerm.getClass()) {
				if (lTerm == rTerm) { // the same (atom)
					// Decomposition
					sigma.composeInRange(genVar, lTerm);
					debug("DEC", debugLevel, debugOut);
					continue;
				}
				if (lHead == rHead) {
					if (lTerm instanceof FunctionApplication) {
						// Decomposition
						FunctionApplication lFnc = (FunctionApplication) lTerm;
						FunctionApplication rFnc = (FunctionApplication) rTerm;
						NominalTerm[] lArgs = lFnc.getArgs();
						NominalTerm[] rArgs = rFnc.getArgs();
						for (int i = 0, n = lArgs.length; i < n; i++) {
							AntiUnifyProblem aup = addEquation(lArgs[i],
									rArgs[i], false);
							lArgs[i] = new Suspension(
									aup.getGeneralizationVar());
						}
						sigma.composeInRange(genVar, lTerm);
						debug("DEC", debugLevel, debugOut);
						continue;
					}
					if (lTerm instanceof Abstraction) {
						// Abstraction
						Abstraction lAbs = (Abstraction) lTerm;
						Abstraction rAbs = (Abstraction) rTerm;
						AntiUnifyProblem aup = addEquation(lAbs.getSubTerm(),
								rAbs.getSubTerm(), false);
						lAbs.setSubTerm(new Suspension(aup
								.getGeneralizationVar()));
						sigma.composeInRange(genVar, lTerm);
						debug("ABS", debugLevel, debugOut);
						continue;
					}
				}
				if (lTerm instanceof Abstraction) {
					// Abstraction
					Abstraction lAbs = (Abstraction) lTerm;
					Abstraction rAbs = (Abstraction) rTerm;
					Atom lBound = lAbs.getBoundAtom();
					Atom rBound = rAbs.getBoundAtom();
					// Test whether the sort matches
					boolean sameSort = false;
					Sort lSort = lBound.getSort();
					Sort rSort = rBound.getSort();
					if (lSort != null && rSort == null) {
						rAbs.getBoundAtom().setSort(lSort);
						sameSort = true;
					} else if (rSort != null && lSort == null) {
						lAbs.getBoundAtom().setSort(rSort);
						sameSort = true;
					} else {
						sameSort = lSort == rSort;
					}
					if (sameSort) {
						Atom c = null;
						for (Atom atom : atoms) {
							if (lTerm.isFresh(atom, nablaIn)
									&& rTerm.isFresh(atom, nablaIn)) {
								c = atom;
								break;
							}
						}
						if (c != null) {
							AntiUnifyProblem aup = addEquation(lAbs
									.getSubTerm().swap(lBound, c), rAbs
									.getSubTerm().swap(rBound, c), false);
							lAbs.setBoundAtom(c);
							lAbs.setSubTerm(new Suspension(aup
									.getGeneralizationVar()));
							sigma.composeInRange(genVar, lTerm);
							debug("ABS", debugLevel, debugOut);
							continue;
						}
					}
				}
			}
			// Solve
			for (Atom atom : atoms) {
				if (lTerm.isFresh(atom, nablaIn)
						&& rTerm.isFresh(atom, nablaIn))
					nablaGen.put(atom, genVar);
			}
			store.add(eq);
			debug("SOL", debugLevel, debugOut);
		}
		debug(debugLevel, debugOut);
		// Merge
		for (int i = store.size() - 2; i >= 0; i--) {
			AntiUnifyProblem eq1 = store.get(i);
			NominalTerm eq1Left = eq1.getLeft();
			NominalTerm eq1Right = eq1.getRight();
			Sort eq1Sort = eq1Left.getSort();
			for (int j = store.size() - 1; j > i; j--) {
				AntiUnifyProblem eq2 = store.get(j);
				NominalTerm eq2Left = eq2.getLeft();
				if (eq1Sort == eq2Left.getSort()) {
					debug(null, debugLevel, debugOut, "Trying to apply"
							+ " MER rule to " + eq1 + " and " + eq2, "");
					permSys.clear().start(atoms, nablaIn);
					permSys.addEquation(eq1Left, eq2Left, true);
					permSys.addEquation(eq1Right, eq2.getRight(), true);
					Permutation pi = permSys.compute(debugLevel, debugOut);
					if (pi != null) {
						debug(null,
								debugLevel,
								debugOut,
								"Equivariance test succeeded."
										+ " Computed permutation = "
										+ pi.toString(true), "");
						Suspension permSusp = new Suspension(
								eq1.getGeneralizationVar(), pi);
						nablaGen.substitute(eq2.getGeneralizationVar(),
								permSusp);
						sigma.composeInRange(eq2.getGeneralizationVar(),
								permSusp);
						store.remove(j);
						debug("MER", debugLevel, debugOut);
						debug(debugLevel, debugOut);
					} else
						debug(null, debugLevel, debugOut,
								"Equivariance test failed.", "");
				}
			}
		}
	}

	/**
	 * Adds an anti-unification problem to the set of problems.
	 */
	public AntiUnifyProblem addEquation(NominalTerm left, NominalTerm right,
			boolean deepCopy) throws ControlledException {
		AntiUnifyProblem aup;
		if (deepCopy)
			aup = problemSet.newEquation(left.deepCopy(), right.deepCopy());
		else
			aup = problemSet.newEquation(left, right);
		problemSet.add(aup);
		return aup;
	}

	private void debug(DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println(" Problem: " + problemSet);
			out.println("   Store: " + store);
			out.println("   Nabla: " + nablaGen);
			out.println("   Sigma: " + sigma);
			out.println();
		}
	}

	private void debug(String rule, DebugLevel debugLevel, PrintStream out,
			String... info) {
		if (debugLevel == DebugLevel.PROGRESS) {
			if (rule != null)
				out.println(rule + " ==> ");
			for (String line : info)
				out.println(line);
		}
	}

	/**
	 * Returns the problem set which is empty after {@linkplain #compute()}
	 * succeeded.
	 */
	public EquationSystem<AntiUnifyProblem> getProblemSet() {
		return problemSet;
	}

	/**
	 * Returns the store where all the disagreements are stored during the
	 * computation. It contains all the information about how to instantiate the
	 * computed generalization to obtain the original terms back. Only contains
	 * information after {@linkplain #compute()} succeeded.
	 */
	public List<AntiUnifyProblem> getStore() {
		return store;
	}

	/**
	 * The substitution sigma contains the generalization after
	 * {@linkplain #compute()} succeeded.
	 */
	public Substitution getSigma() {
		return sigma;
	}

	/**
	 * The freshness context of the generalization after {@linkplain #compute()}
	 * succeeded.
	 */
	public FreshnessCtx getNablaGen() {
		return nablaGen;
	}

	@Override
	public void printString(Writer toPrint) throws IOException {
		sigma.printRanString(toPrint, nablaGen);
	}
}
