/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.algo;

import java.io.PrintStream;
import java.util.Collection;

import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Permutation;
import at.jku.risc.stout.nau.util.ControlledException;

/**
 * This class encapsulates the rule based system {@linkplain EquivarianceSystem}
 * and allows to justify the result by applying the computed permutation and
 * testing alpha equivalence.
 * 
 * @author Alexander Baumgartner
 */
public class Equivariance {
	private EquationSystem<EquivarianceProblem> eqSys;
	private Collection<? extends Atom> atoms;
	private FreshnessCtx nablaIn;

	/**
	 * Creates and initializes the encapsulated rule based system
	 * {@linkplain EquivarianceSystem} with the given equation system, atoms and
	 * freshness context.
	 */
	public Equivariance(EquationSystem<EquivarianceProblem> eqSys,
			Collection<? extends Atom> atoms, FreshnessCtx nablaIn) {
		this.eqSys = eqSys;
		this.nablaIn = nablaIn;
		this.atoms = atoms;
	}

	/**
	 * Calls the rule based system {@linkplain EquivarianceSystem} and tries to
	 * justify the computed result if the second argument of this method is
	 * true.
	 */
	public Permutation compute(NodeFactory factory, boolean justify,
			DebugLevel debugLevel, PrintStream out) throws ControlledException {
		EquivarianceSystem equivSys = new EquivarianceSystem(factory,
				eqSys.deepCopy());
		equivSys.start(atoms, nablaIn);
		Permutation pi = equivSys.compute(debugLevel, out);
		String outString = pi == null ? "Not equivariant!" : pi.toString(true);
		if (debugLevel != DebugLevel.SILENT) {
			out.println("-----------");
			out.println(" Result: " + outString);
			out.println("-----------");
			out.println();
		}
		if (justify) {
			if (pi == null) {
				if (debugLevel != DebugLevel.SILENT)
					out.println("Justification not available!");
			} else {
				if (debugLevel != DebugLevel.SILENT
						&& debugLevel != DebugLevel.SIMPLE)
					out.println("Start justification.");
				for (EquivarianceProblem equivP : eqSys)
					justifyAup(equivP, pi, nablaIn, debugLevel, out);
				if (debugLevel != DebugLevel.SILENT
						&& debugLevel != DebugLevel.SIMPLE)
					out.println("Justification Ok.");
			}
		}
		return pi;
	}

	private void justifyAup(EquivarianceProblem equivP, Permutation pi,
			FreshnessCtx nablaIn, DebugLevel debugLevel, PrintStream out)
			throws ControlledException {
		NominalTerm leftPi = equivP.getLeft().deepCopy().permute(pi);
		if (!leftPi.equivalent(equivP.getRight(), nablaIn))
			throw new JustificationException(nablaIn + " |-/- " + leftPi
					+ " = " + equivP.getRight());
		if (debugLevel != DebugLevel.SILENT && debugLevel != DebugLevel.SIMPLE) {
			if (!pi.isEmpty())
				out.println(pi + " " + equivP.getLeft() + " = " + leftPi);
			out.println(nablaIn + " |-- " + leftPi + " = " + equivP.getRight());
		}
	}

	public EquationSystem<EquivarianceProblem> getEqSys() {
		return eqSys;
	}

	public Collection<? extends Atom> getAtoms() {
		return atoms;
	}

	public FreshnessCtx getNablaIn() {
		return nablaIn;
	}
}
