/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.algo;

import java.io.IOException;
import java.io.PrintStream;
import java.io.StringWriter;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import at.jku.risc.stout.nau.data.EquationSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.data.atom.Abstraction;
import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.FunctionApplication;
import at.jku.risc.stout.nau.data.atom.HasSort;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Permutation;
import at.jku.risc.stout.nau.data.atom.Sort;
import at.jku.risc.stout.nau.data.atom.SortAtom;
import at.jku.risc.stout.nau.data.atom.Suspension;
import at.jku.risc.stout.nau.data.atom.Variable;
import at.jku.risc.stout.nau.util.ControlledException;
import at.jku.risc.stout.nau.util.DataStructureFactory;

/**
 * This class represents a rule based system to compute an equivariance
 * permutation for nominal terms by atom swapping. The nominal terms will be
 * alpha equivalent under the computed permutation. This algorithm is used as
 * auxiliary task in {@linkplain AntiUnifySystem}.<br>
 * The algorithm E is described in the paper:<br>
 * <a href="http://www.risc.jku.at/home/abaumgar/">Alexander Baumgartner</a>, <a
 * href="http://www.risc.jku.at/people/tkutsia/">Temur Kutsia</a>, <a
 * href="http://www.iiia.csic.es/~levy/">Jordi Levy</a> and <a
 * href="http://ima.udg.edu/~villaret/">Mateu Villaret</a>, Nominal
 * Anti-Unificaion
 * 
 * @author Alexander Baumgartner
 */
public class EquivarianceSystem {
	private Set<? extends Atom> atoms;
	private FreshnessCtx nabla;
	private NodeFactory factory;
	private Permutation pi = new Permutation();

	private EquationSystem<EquivarianceProblem> problemSet;
	private Map<Atom, Atom> atomMap = DataStructureFactory.$.newMap();
	private List<Atom> atomList = DataStructureFactory.$.newList();

	public EquivarianceSystem(NodeFactory factory) {
		this(factory, new EquationSystem<EquivarianceProblem>() {
			@Override
			public EquivarianceProblem newEquation(NominalTerm t, NominalTerm s) {
				return new EquivarianceProblem(t, s);
			}
		});
	}

	public EquivarianceSystem(NodeFactory factory,
			EquationSystem<EquivarianceProblem> problemSet) {
		this.factory = factory;
		this.problemSet = problemSet;
	}

	/**
	 * Initializes the permutation computation.
	 */
	public void start(Collection<? extends Atom> atoms, FreshnessCtx nablaIn) {
		this.atoms = DataStructureFactory.$.newSet(atoms);
		this.nabla = nablaIn;
	}

	public EquivarianceSystem clear() {
		this.pi.clear();
		this.problemSet.clear();
		this.atomMap.clear();
		this.atomList.clear();
		return this;
	}

	/**
	 * Adds a matching problem to the problem set.
	 */
	public void addEquation(NominalTerm eq1Left, NominalTerm eq2Left,
			boolean deepCopy) {
		if (deepCopy)
			problemSet.add(problemSet.newEquation(eq1Left.deepCopy(),
					eq2Left.deepCopy()));
		else
			problemSet.add(problemSet.newEquation(eq1Left, eq2Left));
	}

	/**
	 * Computes the equivariance permutation silently.
	 * 
	 * @see #compute(DebugLevel, PrintStream)
	 */
	public Permutation compute() {
		try {
			return compute(DebugLevel.SILENT, null);
		} catch (ControlledException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Computes an equivariance permutation for given equivariance problems (see
	 * {@linkplain #addEquation(NominalTerm, NominalTerm, boolean)}) and given
	 * atoms/nabla (see {@linkplain #start(Collection, FreshnessCtx)}). Returns
	 * null if a clash appears, otherwise a {@linkplain Permutation} will be
	 * returned.
	 */
	public Permutation compute(DebugLevel debugLevel, PrintStream out)
			throws ControlledException {
		// Phase 1
		debug(null, debugLevel, out,
				"Starting phase 1 (simplification) of equivariance algorithm",
				"");
		debug(debugLevel, out);
		while (!problemSet.isEmpty()) {
			EquivarianceProblem eq = problemSet.popLast();
			NominalTerm lTerm = eq.getLeft();
			NominalTerm rTerm = eq.getRight();
			HasSort<? extends Sort> lHead = lTerm.getHead();
			HasSort<? extends Sort> rHead = rTerm.getHead();

			if (lTerm.getClass() != rTerm.getClass()
					|| lHead.getSort() != rHead.getSort()) {
				debug("HEAD-CLASH", debugLevel, out, "No rule applicable to "
						+ eq);
				return null; // No rule applicable
			}

			if (lTerm instanceof Abstraction) {
				// Alpha Equivalence
				Abstraction lAbs = (Abstraction) lTerm;
				Abstraction rAbs = (Abstraction) rTerm;
				NominalTerm lSub = lAbs.getSubTerm();
				NominalTerm rSub = rAbs.getSubTerm();
				SortAtom sort = lAbs.getBoundAtom().getSort();
				Atom freshAtom = factory.obtainFreshAtom(sort);
				eq.setLeft(lSub.swap(lAbs.getBoundAtom(), freshAtom));
				eq.setRight(rSub.swap(rAbs.getBoundAtom(), freshAtom));
				problemSet.add(eq);
				debug("ALP-E", debugLevel, out);
			} else if (lTerm instanceof FunctionApplication) {
				// Decomposition
				FunctionApplication lFnc = (FunctionApplication) lTerm;
				FunctionApplication rFnc = (FunctionApplication) rTerm;
				if (lFnc.getFncSymb() != rFnc.getFncSymb()) {
					debug("HEAD-CLASH", debugLevel, out,
							"No rule applicable to " + eq);
					return null; // No rule applicable
				}
				NominalTerm[] lArgs = lFnc.getArgs();
				NominalTerm[] rArgs = rFnc.getArgs();
				for (int i = lArgs.length - 1; i >= 0; i--)
					addEquation(lArgs[i], rArgs[i], false);
				debug("DEC-E", debugLevel, out);
			} else if (lTerm instanceof Suspension) {
				// Resolve suspension
				Suspension lSusp = (Suspension) lTerm;
				Suspension rSusp = (Suspension) rTerm;
				Variable var = lSusp.getVar();
				if (var != rSusp.getVar()) {
					debug("HEAD-CLASH", debugLevel, out,
							"No rule applicable to " + eq);
					return null; // No rule applicable
				}
				Permutation lPerm = lSusp.getPerm();
				Permutation rPerm = rSusp.getPerm();
				debug("SUS-E", debugLevel, out);
				for (Atom atom : atoms) {
					if (!nabla.contains(atom, var)) {
						if (!putAtom(debugLevel, out, lPerm.permute(atom),
								rPerm.permute(atom)))
							return null;

					}
				}
			} else {
				if (!putAtom(debugLevel, out, (Atom) lTerm, (Atom) rTerm))
					return null; // Shift atom to atomMap and detect clash
				continue;
			}
			debug(debugLevel, out);
		}
		// Phase 2
		debug(null, debugLevel, out,
				"Starting phase 2 (permutation) of equivariance algorithm", "");
		for (Atom lAtom : atomList) {
			Atom rAtom = atomMap.remove(lAtom);
			lAtom = pi.permute(lAtom);
			if (lAtom == rAtom) {
				atoms.remove(lAtom);
				debug("REM-E", debugLevel, out);
			} else if (atoms.contains(lAtom) && atoms.remove(rAtom)) {
				pi.addSwappingHead(lAtom, rAtom);
				debug("SOL-E", debugLevel, out);
			} else {
				debug("PERM-CLASH", debugLevel, out,
						"Permutation application not possible to " + lAtom
								+ EquivarianceProblem.eqSeparator + rAtom);
				return null; // No rule applicable
			}
			debug(debugLevel, out);
		}
		return pi;
	}

	public boolean putAtom(DebugLevel debugLevel, PrintStream out, Atom lAtom,
			Atom rAtom) {
		Atom oldVal = atomMap.put(lAtom, rAtom);
		if (oldVal == null) {
			atomList.add(lAtom);
		} else if (oldVal != rAtom) {
			debug("ATOM-CLASH", debugLevel, out, "Swapping conflict: " + lAtom
					+ EquivarianceProblem.eqSeparator + oldVal + "; " + lAtom
					+ EquivarianceProblem.eqSeparator + rAtom);
			return false;
		}
		return true;
	}

	private void debug(DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("    Atoms: " + atoms);
			out.println("Equations: " + problemToString());
			out.println("       Pi: " + pi.toString(true));
			out.println();
		}
	}

	private void debug(String rule, DebugLevel debugLevel, PrintStream out,
			String... info) {
		if (debugLevel == DebugLevel.PROGRESS) {
			if (rule != null)
				out.println(rule + " ==> ");
			for (String line : info)
				out.println(line);
		}
	}

	private String problemToString() {
		StringWriter sw = new StringWriter();
		try {
			problemSet.printString(sw);
			int i = problemSet.size();
			for (Entry<Atom, Atom> entry : atomMap.entrySet()) {
				EquationSystem.writePrefix(sw, i);
				EquivarianceProblem.printString(sw, entry.getKey(),
						entry.getValue());
				i++;
			}
		} catch (IOException e) {
			return e.toString();
		}
		return sw.toString();
	}

	@Override
	public String toString() {
		return pi.toString();
	}
}
