/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.algo;

import java.io.IOException;
import java.io.Writer;
import java.util.Map;
import java.util.Map.Entry;

import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.NominalPair;
import at.jku.risc.stout.nau.data.atom.MalformedSortException;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Variable;
import at.jku.risc.stout.nau.util.DataStructureFactory;

/**
 * This class represents a substitution, which is a mapping from variables to
 * nominal terms.<br>
 * It is used inside the rule based system {@linkplain AntiUnifySystem} to
 * compute generalizations for given {@linkplain AntiUnifyProblem}s. In this
 * case, the left hand side of the mapping is the generalization variable of an
 * anti-unification problem (AUP) and the right hand side is the generalization
 * computed so far.
 * 
 * @author Alexander Baumgartner
 */
public class Substitution implements Cloneable {
	private Map<Variable, NominalTerm> mapping = DataStructureFactory.$
			.newMap();
	/**
	 * Default = "{"
	 */
	public static String SIGMA_START = "{";
	/**
	 * Default = " -> "
	 */
	public static String SIGMA_MAPTO = " -> ";
	/**
	 * Default = "} "
	 */
	public static String SIGMA_END = "} ";
	/**
	 * Default = "; "
	 */
	public static String RAN_PRINT_SEPARATOR = "; ";

	/**
	 * Add a new mapping of the form: Variable -&gt; Term.
	 */
	public void put(Variable fromVar, NominalTerm toTerm) {
		mapping.put(fromVar, toTerm);
	}

	/**
	 * Substitution composition which does not add new variables to the mapping.
	 * (This is useful for generalization computation of AUPs.)
	 */
	public void composeInRange(Variable fromVar, NominalTerm toTerm)
			throws MalformedSortException {
		for (Entry<Variable, NominalTerm> e : mapping.entrySet())
			e.setValue(e.getValue().substitute(fromVar, toTerm));
	}

	/**
	 * Returns the associated {@linkplain NominalTerm} for a given
	 * {@linkplain Variable} or null if no mapping exists.
	 */
	public NominalTerm get(Variable var) {
		return mapping.get(var);
	}

	/**
	 * Removes all the mappings from a substitution.
	 */
	public void clear() {
		mapping.clear();
	}

	@Override
	public Substitution clone() {
		try {
			Substitution clone = (Substitution) super.clone();
			clone.mapping = DataStructureFactory.$.newMap(mapping);
			return clone;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}

	public Map<Variable, NominalTerm> getMapping() {
		return mapping;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (Entry<Variable, NominalTerm> e : mapping.entrySet()) {
			sb.append(SIGMA_START);
			sb.append(e.getKey().getName());
			sb.append(SIGMA_MAPTO);
			sb.append(String.valueOf(e.getValue()));
			sb.append(SIGMA_END);
		}
		return sb.toString();
	}

	/**
	 * Only prints the range of the mapping, which is actually useful to display
	 * a computed generalization of an AUP without the generalization variable.
	 */
	public void printRanString(Writer writer, FreshnessCtx nabla)
			throws IOException {
		boolean first = true;
		for (Entry<Variable, NominalTerm> e : mapping.entrySet()) {
			if (first)
				first = false;
			else
				writer.write(RAN_PRINT_SEPARATOR);
			NominalPair.printPair(writer, nabla, e.getValue());
		}
	}
}