/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data;

import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import at.jku.risc.stout.nau.data.atom.Atom;
import at.jku.risc.stout.nau.data.atom.FunctionSymbol;
import at.jku.risc.stout.nau.data.atom.HasSort;
import at.jku.risc.stout.nau.data.atom.MalformedSortException;
import at.jku.risc.stout.nau.data.atom.NominalTerm;
import at.jku.risc.stout.nau.data.atom.Sort;
import at.jku.risc.stout.nau.data.atom.SortAtom;
import at.jku.risc.stout.nau.data.atom.SortData;
import at.jku.risc.stout.nau.data.atom.Variable;
import at.jku.risc.stout.nau.util.ControlledException;
import at.jku.risc.stout.nau.util.DataStructureFactory;

/**
 * This factory is used to instantiate new {@linkplain NominalTerm}s, subterms
 * and atomic term blocks.
 * 
 * @author Alexander Baumgartner
 */
public class NodeFactory {
	private Map<Class<?>, Map<List<Object>, ?>> objects = DataStructureFactory.$
			.newMap();
	private List<Object> argsTmp = DataStructureFactory.$.newList(2);
	private String prefixFreshAtom = "&";
	private String suffixFreshAtom = "";
	private int uniqueAtomCount;
	private String prefixFreshVar = "%";
	private String suffixFreshVar = "";
	private int uniqueVarCount;
	/**
	 * Class to use for atom instantiation. There must be a constructor with one
	 * argument &quot;String name&quot;!
	 */
	public Class<? extends Atom> classAtom = Atom.class;
	/**
	 * Class to use for variable instantiation. There must be a constructor with
	 * one argument &quot;String name&quot;!
	 */
	public Class<? extends Variable> classVariable = Variable.class;
	/**
	 * Class to use for function-symbol instantiation. There must be a
	 * constructor with one argument &quot;String name&quot;!
	 */
	public Class<? extends FunctionSymbol> classFncSymb = FunctionSymbol.class;
	/**
	 * Class to use for instantiating a sort of atom. There must be a
	 * constructor with one argument &quot;String name&quot;!
	 */
	public Class<? extends SortAtom> classSortAtom = SortAtom.class;
	/**
	 * Class to use for instantiating a sort of data. There must be a
	 * constructor with one argument &quot;String name&quot;!
	 */
	public Class<? extends SortData> classSortData = SortData.class;

	public NodeFactory() {
	}

	public NodeFactory(String prefixFreshAtom, String prefixFreshVar) {
		this.prefixFreshAtom = prefixFreshAtom;
		this.prefixFreshVar = prefixFreshVar;
	}

	/**
	 * Instantiates an untyped atom.
	 */
	public Atom newAtom(String name) throws FactoryInstantiationException {
		try {
			return newAtom(name, (SortAtom) null);
		} catch (MalformedSortException e) {
			throw new IllegalStateException(
					"Typing error should not happen here!", e);
		}
	}

	/**
	 * Instantiates an atom of the specified sort.
	 */
	public Atom newAtom(String name, String sort)
			throws FactoryInstantiationException, MalformedSortException {
		if (sort == null)
			return newAtom(name, (SortAtom) null);
		return newAtom(name, newSort(classSortAtom, sort));
	}

	/**
	 * Instantiates an atom of the specified sort.
	 */
	public Atom newAtom(String name, SortAtom sort)
			throws FactoryInstantiationException, MalformedSortException {
		Atom ret = uniqueInstance(classAtom, name.intern());
		return checkSort(name, (Sort) sort, ret);
	}

	public <T> T getExists(Class<T> type, String name) {
		return getInstance(getInstancePool(type), name.intern());
	}

	public <T> T getExists(Class<T> type, Object... args) {
		return getInstance(getInstancePool(type), args);
	}

	/**
	 * Instantiates an untyped variable.
	 */
	public Variable newVariable(String name)
			throws FactoryInstantiationException {
		try {
			return newVariable(name, null);
		} catch (MalformedSortException e) {
			throw new IllegalStateException(
					"Typing error should not happen here!", e);
		}
	}

	/**
	 * Instantiates a variable of the specified sort.
	 */
	public Variable newVarAtom(String name, String sort)
			throws FactoryInstantiationException, MalformedSortException {
		if (sort == null)
			return newVariable(name, (SortAtom) null);
		return newVariable(name, newSort(classSortAtom, sort));
	}

	/**
	 * Instantiates a variable of the specified sort.
	 */
	public Variable newVarData(String name, String sort)
			throws FactoryInstantiationException, MalformedSortException {
		if (sort == null)
			return newVariable(name, (SortData) null);
		return newVariable(name, newSort(classSortData, sort));
	}

	/**
	 * Instantiates a variable of the specified sort.
	 */
	public Variable newVariable(String name, Sort sort)
			throws FactoryInstantiationException, MalformedSortException {
		Variable ret = uniqueInstance(classVariable, name.intern());
		return checkSort(name, sort, ret);
	}

	/**
	 * Instantiates an untyped function symbol.
	 */
	public FunctionSymbol newFunction(String name)
			throws FactoryInstantiationException {
		try {
			return newFunction(name, null, (SortData) null);
		} catch (MalformedSortException e) {
			throw new IllegalStateException(
					"Typing error should not happen here!", e);
		}
	}

	/**
	 * Instantiates a function symbol of the specified arity.
	 */
	public FunctionSymbol newFunction(String name, Sort[] sortArgs,
			String sortRet) throws FactoryInstantiationException,
			MalformedSortException {
		if (sortRet == null)
			return newFunction(name, sortArgs, (SortData) null);
		return newFunction(name, sortArgs, newSort(classSortData, sortRet));
	}

	/**
	 * Instantiates a function symbol of the specified arity.
	 */
	public FunctionSymbol newFunction(String name, Sort[] sortArgs,
			SortData sortRet) throws FactoryInstantiationException,
			MalformedSortException {
		FunctionSymbol ret = uniqueInstance(classFncSymb, name.intern());
		if (ret.getSortArgs() == null) {
			if (sortArgs != null)
				ret.setSortArgs(sortArgs);
		} else if (sortArgs != null
				&& !Arrays.equals(sortArgs, ret.getSortArgs())) {
			throw new MalformedSortException("FunctionSymbol '" + name
					+ "' has two different sorts");
		}
		return checkSort(name, sortRet, ret);
	}

	private <S extends Sort, T extends HasSort<S>> T checkSort(String name,
			S sort, T ret) throws MalformedSortException {
		if (ret.getSort() == null) {
			if (sort != null)
				ret.setSort(sort);
		} else if (sort != null && sort != ret.getSort()) {
			throw new MalformedSortException("Symbol '" + name
					+ "' has two different sorts");
		}
		return ret;
	}

	/**
	 * Returns a sort of atom by name.
	 */
	public <T extends Sort> T newSort(Class<T> sortType, String name)
			throws FactoryInstantiationException {
		return uniqueInstance(sortType, name.intern());
	}

	/**
	 * Returns a unique instance of the specified class for some given
	 * constructor arguments. An accessible constructor with the specified
	 * argument types must exist for the specified type!
	 * 
	 * @return Unique instance for given type and arguments.
	 * @throws FactoryInstantiationException
	 *             If no corresponding constructor exists.
	 */
	public <T> T uniqueInstance(Class<T> type, Object... args)
			throws FactoryInstantiationException {
		Map<List<Object>, T> instances = getInstancePool(type);
		T instance = getInstance(instances, args);
		if (instance == null) {
			@SuppressWarnings("unchecked")
			Constructor<T>[] cs = (Constructor<T>[]) type
					.getDeclaredConstructors();
			Constructor<T> c = null;
			for (int i = 0; c == null && i < cs.length; i++) {
				c = cs[i];
				Class<?>[] types = c.getParameterTypes();
				if ((c.getModifiers() & (Modifier.PROTECTED | Modifier.PUBLIC)) == 0
						|| types.length != args.length) {
					c = null;
					continue;
				}
				for (int j = 0; j < args.length; j++) {
					Object arg = args[j];
					if (arg != null && arg.getClass() != types[j]) {
						c = null;
						break;
					}
				}
			}
			if (c == null)
				throw new FactoryInstantiationException(
						"No matching constructor found!");
			try {
				instance = c.newInstance(args);
			} catch (Exception e) {
				throw new FactoryInstantiationException(e);
			}
			instances.put(argsTmp, instance);
			argsTmp = DataStructureFactory.$.newList(2);
		}
		return instance;
	}

	private <T> T getInstance(Map<List<Object>, T> instancePool, Object... args) {
		argsTmp.clear();
		for (int i = 0; i < args.length; i++)
			argsTmp.add(args[i]);
		T instance = instancePool.get(argsTmp);
		return instance;
	}

	public <T> Collection<T> getAllByType(Class<T> type) {
		return getInstancePool(type).values();
	}

	private <T> Map<List<Object>, T> getInstancePool(Class<T> type) {
		@SuppressWarnings("unchecked")
		Map<List<Object>, T> instances = (Map<List<Object>, T>) objects
				.get(type);
		if (instances == null) {
			instances = DataStructureFactory.$.newMap();
			objects.put(type, instances);
		}
		return instances;
	}

	public class FactoryInstantiationException extends ControlledException {
		private static final long serialVersionUID = -4997440623071173164L;

		FactoryInstantiationException(Throwable cause) {
			super("Object creation failed!", cause);
		}

		FactoryInstantiationException(String msg) {
			super(msg);
		}
	}

	public String obtainUniqueAtomName() {
		return prefixFreshAtom + (++uniqueAtomCount) + suffixFreshAtom;
	}

	public String obtainUniqueVarName() {
		return prefixFreshVar + (++uniqueVarCount) + suffixFreshVar;
	}

	public Variable obtainFreshVar(Sort sort) {
		try {
			return newVariable(obtainUniqueVarName(), sort);
		} catch (ControlledException e) {
			throw new InstantiationError(e.getMessage());
		}
	}

	public Atom obtainFreshAtom(SortAtom sort) {
		try {
			return newAtom(obtainUniqueAtomName(), sort);
		} catch (ControlledException e) {
			throw new InstantiationError(e.getMessage());
		}
	}

	public String getPrefixFreshAtom() {
		return prefixFreshAtom;
	}

	public void setPrefixFreshAtom(String prefixFreshAtom) {
		this.prefixFreshAtom = prefixFreshAtom;
	}

	public String getSuffixFreshAtom() {
		return suffixFreshAtom;
	}

	public void setSuffixFreshAtom(String suffixFreshAtom) {
		this.suffixFreshAtom = suffixFreshAtom;
	}

	public String getPrefixFreshVar() {
		return prefixFreshVar;
	}

	public void setPrefixFreshVar(String prefixFreshVar) {
		this.prefixFreshVar = prefixFreshVar;
	}

	public String getSuffixFreshVar() {
		return suffixFreshVar;
	}

	public void setSuffixFreshVar(String suffixFreshVar) {
		this.suffixFreshVar = suffixFreshVar;
	}

	public void resetCounter() {
		uniqueAtomCount = 0;
		uniqueVarCount = 0;
	}
}
