/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data.atom;

import java.io.IOException;
import java.io.Writer;
import java.util.Set;

import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.InputParser;

/**
 * An abstraction is a {@linkplain NominalTerm} which consists of a bound atom
 * and a subterm.
 * 
 * @author Alexander Baumgartner
 */
public class Abstraction extends NominalTerm {
	private Atom boundAtom;
	private NominalTerm subTerm;

	public Abstraction(Atom boundAtom, NominalTerm subTerm) {
		this.boundAtom = boundAtom;
		this.subTerm = subTerm;
	}

	@Override
	public boolean equals(Object other) {
		if (other instanceof Abstraction) {
			Abstraction oAbs = (Abstraction) other;
			// Atom instances are unique
			return oAbs.boundAtom == boundAtom && oAbs.subTerm.equals(subTerm);
		}
		return false;
	}

	public Atom getBoundAtom() {
		return boundAtom;
	}

	public void setBoundAtom(Atom boundAtom) {
		this.boundAtom = boundAtom;
	}

	public NominalTerm getSubTerm() {
		return subTerm;
	}

	public void setSubTerm(NominalTerm subTerm) {
		this.subTerm = subTerm;
	}

	@Override
	public Sort getSort() {
		return subTerm.getSort();
	}

	@Override
	public void setSort(Sort sort) throws MalformedSortException {
		subTerm.setSort(sort);
	}

	@Override
	public void printString(Writer toPrint) throws IOException {
		toPrint.write(boundAtom.toString());
		toPrint.write(InputParser.cp_abstraction);
		subTerm.printString(toPrint);
	}

	@Override
	public NominalTerm swap(Atom a1, Atom a2) throws MalformedSortException {
		boundAtom = boundAtom.swap(a1, a2);
		subTerm = subTerm.swap(a1, a2);
		return this;
	}

	@Override
	protected boolean propagate(TraverseCallBack<NominalTerm> callBack) {
		return subTerm.traverse(callBack);
	}

	@Override
	public boolean isFresh(Atom atom, FreshnessCtx nabla) {
		if (atom == boundAtom) // #-abst-1
			return true;
		return subTerm.isFresh(atom, nabla); // #-abst-2
	}

	@Override
	public Abstraction deepCopy() {
		return new Abstraction(boundAtom, subTerm.deepCopy());
	}

	@Override
	public NominalTerm substitute(Variable fromVar, NominalTerm toTerm)
			throws MalformedSortException {
		subTerm = subTerm.substitute(fromVar, toTerm);
		return this;
	}

	@Override
	public Abstraction permute(Permutation perm) {
		boundAtom = boundAtom.permute(perm);
		subTerm = subTerm.permute(perm);
		return this;
	}

	@Override
	public HasSort<Sort> getHead() {
		return boundAtom;
	}

	@Override
	public void collectAtoms(Set<Atom> atoms) {
		atoms.add(boundAtom);
		subTerm.collectAtoms(atoms);
	}
}
