/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data.atom;

import java.io.IOException;
import java.io.Writer;
import java.util.Set;

import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.InputParser;
import at.jku.risc.stout.nau.data.NodeFactory;

/**
 * An {@linkplain Atom} is a {@linkplain NominalTerm}, consisting of the atom
 * name and an optional origin name. The origin name may be used to store the
 * name of a bound atom before renaming it. Atoms may be typed by a sort of
 * atom.
 * 
 * @author Alexander Baumgartner
 */
public class Atom extends NominalTerm {
	private String name;
	private String nameOrigin;
	private SortAtom sort;

	/**
	 * Instantiates an atom of the specified name. Use the {@link NodeFactory}
	 * to create new {@link Atom}s! The use of {@link NodeFactory} guarantees
	 * the uniqueness of equal {@link Atom}s. This uniqueness enables very fast
	 * equality testing, hash-code generation, comparison of atoms,... and it
	 * keeps the memory overhead as small as possible.
	 */
	public Atom(String name) {
		this(name, null);
	}

	/**
	 * Instantiates an atom of the specified name. Use the {@link NodeFactory}
	 * to create new {@link Atom}s! The use of {@link NodeFactory} guarantees
	 * the uniqueness of equal {@link Atom}s. This uniqueness enables very fast
	 * equality testing, hash-code generation, comparison of atoms,... and it
	 * keeps the memory overhead as small as possible.
	 */
	public Atom(String name, String nameOrigin) {
		this.name = name;
		this.nameOrigin = nameOrigin;
	}

	public String getName() {
		return name;
	}

	public String getNameOrigin() {
		return nameOrigin;
	}

	public void setNameOrigin(String nameOrigin) {
		this.nameOrigin = nameOrigin;
	}

	@Override
	public SortAtom getSort() {
		return sort;
	}

	@Override
	public void setSort(Sort sort) throws MalformedSortException {
		if (sort != null && !(sort instanceof SortAtom))
			throw new MalformedSortException(
					"Only SortAtom is allowed for atoms");
		this.sort = (SortAtom) sort;
	}

	@Override
	public void printString(Writer toPrint) throws IOException {
		toPrint.write(name);
		if (sort != null) {
			toPrint.write(InputParser.cp_typingSeparator);
			toPrint.write(sort.toString());
		}
	}

	@Override
	public Atom swap(Atom a1, Atom a2) {
		if (this == a1)
			return a2;
		if (this == a2)
			return a1;
		return this;
	}

	@Override
	public boolean isFresh(Atom atom, FreshnessCtx nabla) {
		return this != atom; // not same object
	}

	@Override
	public NominalTerm deepCopy() {
		return this;
	}

	@Override
	public NominalTerm substitute(Variable fromVar, NominalTerm toTerm) {
		return this;
	}

	@Override
	public Atom permute(Permutation perm) {
		return perm.permute(this);
	}

	@Override
	public HasSort<Sort> getHead() {
		return this;
	}

	@Override
	public void collectAtoms(Set<Atom> atoms) {
		atoms.add(this);
	}
}
