/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data.atom;

import java.io.IOException;
import java.io.Writer;
import java.util.List;
import java.util.Set;

import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.InputParser;
import at.jku.risc.stout.nau.util.DataStructureFactory;

/**
 * A {@linkplain FunctionApplication} is a {@linkplain NominalTerm} consisting
 * of a unique {@linkplain FunctionSymbol} and an array of arguments.
 * 
 * @author Alexander Baumgartner
 */
public class FunctionApplication extends NominalTerm {
	private FunctionSymbol fncSymb;
	private NominalTerm[] args;

	public static final NominalTerm[] emptyArgs = new NominalTerm[0];

	/**
	 * Instantiates a constant without arguments.
	 */
	public FunctionApplication(FunctionSymbol fncSymb)
			throws MalformedSortException {
		this(fncSymb, emptyArgs);
	}

	/**
	 * Instantiates a function application with the given arguments. Use
	 * {@link #FunctionApplication(FunctionSymbol, NominalTerm[])} if possible.
	 */
	public FunctionApplication(FunctionSymbol fncSymb, List<NominalTerm> args)
			throws MalformedSortException {
		this(fncSymb, args == null ? emptyArgs : args.toArray(emptyArgs));
	}

	/**
	 * Instantiates a function application with the given arguments.
	 */
	public FunctionApplication(FunctionSymbol fncSymb, NominalTerm[] args)
			throws MalformedSortException {
		if (args == null || args.length == 0)
			args = emptyArgs;
		this.fncSymb = fncSymb;
		this.args = args;
		int arity = args.length;
		if (fncSymb.isArityUnknown())
			fncSymb.setArity(arity);
		else if (fncSymb.getArity() != arity)
			throw new MalformedSortException(
					"Function application must be in eta-long form");
		if (fncSymb.getSortArgs() == null) {
			if (arity > 0 && args[0].getSort() != null) {
				// Basic trivial type inference
				Sort[] sortArgs = new Sort[arity];
				for (int i = 0; i < arity; i++)
					sortArgs[i] = args[i].getSort();
				fncSymb.setSortArgs(sortArgs);
			}
		} else {
			for (int i = arity - 1; i >= 0; i--) {
				NominalTerm argI = args[i];
				Sort argSort = argI.getSort();
				Sort symbSort = fncSymb.getSortArgs()[i];
				if (argSort == null) {
					if (symbSort != null)
						argI.setSort(symbSort);
				} else if (symbSort == null) {
					fncSymb.getSortArgs()[i] = argSort;
				} else if (symbSort != argSort)
					throw new MalformedSortException(
							"Argument sorts don't match symbol sort");
			}
		}
	}

	@Override
	public boolean equals(Object other) {
		if (other instanceof FunctionApplication) {
			FunctionApplication oApp = (FunctionApplication) other;
			// Function instances are unique
			return oApp.fncSymb == fncSymb && oApp.args.equals(args);
		}
		return false;
	}

	/**
	 * Same as {@linkplain FunctionApplication#getFncSymb()}.getSortRet()
	 */
	@Override
	public Sort getSort() {
		return fncSymb.getSort();
	}

	public FunctionSymbol getFncSymb() {
		return fncSymb;
	}

	public NominalTerm[] getArgs() {
		return args;
	}

	@Override
	public void setSort(Sort sort) throws MalformedSortException {
		if (!(sort instanceof SortData) && sort != null)
			throw new MalformedSortException(
					"Only SortData is allowed for return type of functions");
		fncSymb.setSort((SortData) sort);
	}

	@Override
	public void printString(Writer writer) throws IOException {
		writer.write(fncSymb.toString());
		int len = args.length;
		if (len > 0) {
			writer.write(InputParser.cp_argsStart);
			args[0].printString(writer);
			for (int i = 1; i < len; i++) {
				writer.write(InputParser.cp_argsSeparator);
				writer.write(' ');
				args[i].printString(writer);
			}
			writer.write(InputParser.cp_argsEnd);
		}
	}

	@Override
	public NominalTerm swap(Atom a1, Atom a2) throws MalformedSortException {
		for (int i = args.length - 1; i >= 0; i--)
			args[i] = args[i].swap(a1, a2);
		return this;
	}

	@Override
	protected boolean propagate(TraverseCallBack<NominalTerm> callBack) {
		for (int i = 0, n = args.length; i < n; i++)
			if (args[i].traverse(callBack))
				return true;
		return false;
	}

	@Override
	public boolean isFresh(Atom atom, FreshnessCtx nabla) {
		for (NominalTerm arg : args)
			if (!arg.isFresh(atom, nabla))
				return false;
		return true;
	}

	@Override
	public FunctionApplication deepCopy() {
		try {
			if (args.length == 0)
				return new FunctionApplication(fncSymb, emptyArgs);
			List<NominalTerm> tmp = DataStructureFactory.$.newList(args.length);
			for (NominalTerm arg : args)
				tmp.add(arg.deepCopy());
			return new FunctionApplication(fncSymb, tmp);
		} catch (MalformedSortException e) {
			throw new IllegalStateException(e);
		}
	}

	@Override
	public NominalTerm substitute(Variable fromVar, NominalTerm toTerm)
			throws MalformedSortException {
		for (int i = args.length - 1; i >= 0; i--)
			args[i] = args[i].substitute(fromVar, toTerm);
		return this;
	}

	@Override
	public NominalTerm permute(Permutation perm) {
		for (int i = args.length - 1; i >= 0; i--)
			args[i] = args[i].permute(perm);
		return this;
	}

	@Override
	public HasSort<SortData> getHead() {
		return fncSymb;
	}

	@Override
	public void collectAtoms(Set<Atom> atoms) {
		for (NominalTerm arg : args)
			arg.collectAtoms(atoms);
	}
}
