/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data.atom;

import at.jku.risc.stout.nau.data.InputParser;
import at.jku.risc.stout.nau.data.NodeFactory;

/**
 * A {@linkplain FunctionSymbol} has a unique name and may be typed by a sort of
 * data. The arity and the sorts of its arguments can be specified explicitly.
 * If the arity is unknown, then the value is set to -1 (see
 * {@linkplain #isArityUnknown()}).
 * 
 * @author Alexander Baumgartner
 */
public class FunctionSymbol implements HasSort<SortData> {
	private String name;
	private int arity = -1; // unknown
	private SortData sortRet;
	private Sort[] sortArgs;

	/**
	 * Instantiates a function symbol of the specified name. Use the
	 * {@link NodeFactory} to create new {@link FunctionSymbol}s! The use of
	 * {@link NodeFactory} guarantees the uniqueness of equal
	 * {@link FunctionSymbol}s. This uniqueness enables very fast equality
	 * testing, hash-code generation, comparison of symbols,... and it keeps the
	 * memory overhead as small as possible.
	 */
	public FunctionSymbol(String name) {
		this.name = name;
	}

	@Override
	public SortData getSort() {
		return sortRet;
	}

	@Override
	public void setSort(SortData sortRet) {
		this.sortRet = sortRet;
	}

	public Sort[] getSortArgs() {
		return sortArgs;
	}

	/**
	 * Defines the sorts of the arguments and checks whether it is consistent
	 * with the defined arity.
	 */
	public void setSortArgs(Sort[] sortArgs) throws MalformedSortException {
		if (sortArgs != null) {
			if (isArityUnknown())
				arity = sortArgs.length;
			else if (sortArgs.length != arity)
				throw new MalformedSortException(
						"Sort does not match with arity");
		}
		this.sortArgs = sortArgs;
	}

	public String getName() {
		return name;
	}

	/**
	 * Returns the arity which has been set by the user or could be inferred by
	 * the type system.
	 */
	public int getArity() {
		return arity;
	}

	/**
	 * Returns true if the arity has not been set and could not be inferred so
	 * far.
	 */
	public boolean isArityUnknown() {
		return arity == -1;
	}

	/**
	 * Sets the arity and checks whether it is consistent with the defined sorts
	 * of the arguments.
	 */
	public void setArity(int arity) throws MalformedSortException {
		if (sortArgs != null && sortArgs.length != arity)
			throw new MalformedSortException("Sort does not match with arity");
		this.arity = arity;
	}

	@Override
	public String toString() {
		if (sortRet != null) {
			StringBuilder ret = new StringBuilder();
			ret.append(name).appendCodePoint(InputParser.cp_typingSeparator);
			if (sortArgs != null)
				for (Sort sort : sortArgs)
					ret.append(sort).appendCodePoint(
							InputParser.cp_typeConstructor);
			ret.append(sortRet);
			return ret.toString();
		}
		return name;
	}
}
