/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data.atom;

import java.util.Set;

import at.jku.risc.stout.nau.algo.EquivarianceSystem;
import at.jku.risc.stout.nau.data.FreshnessCtx;
import at.jku.risc.stout.nau.data.NodeFactory;
import at.jku.risc.stout.nau.util.DataStructureFactory;
import at.jku.risc.stout.nau.util.DeepCopy;
import at.jku.risc.stout.nau.util.Printable;
import at.jku.risc.stout.nau.util.Traversable;

/**
 * This is the base class for nominal terms like {@linkplain Abstraction},
 * {@linkplain FunctionApplication}, {@linkplain Atom}, {@linkplain Suspension}.
 * 
 * @author Alexander Baumgartner
 */
public abstract class NominalTerm extends Printable implements
		DeepCopy<NominalTerm>, HasSort<Sort>, Traversable<NominalTerm> {

	public abstract NominalTerm swap(Atom a1, Atom a2)
			throws MalformedSortException;

	/**
	 * Alpha-equivalence test of two nominal terms respecting nabla.
	 */
	public boolean equivalent(NominalTerm otherTerm, FreshnessCtx nabla) {
		EquivarianceSystem ps = new EquivarianceSystem(new NodeFactory());
		Set<Atom> atoms = DataStructureFactory.$.newSet();
		collectAtoms(atoms);
		otherTerm.collectAtoms(atoms);
		ps.start(atoms, nabla);
		ps.addEquation(this, otherTerm, true);
		Permutation pi = ps.compute();
		return pi != null && pi.isEmpty();
	}

	public abstract void collectAtoms(Set<Atom> atoms);

	public abstract boolean isFresh(Atom atom, FreshnessCtx nabla);

	public abstract HasSort<? extends Sort> getHead();

	public abstract NominalTerm substitute(Variable fromVar, NominalTerm toTerm)
			throws MalformedSortException;

	public abstract NominalTerm permute(Permutation perm);

	/**
	 * Traverses the term tree and executes the callback function at every
	 * subterm.
	 * 
	 * @see Traversable
	 * @see at.jku.risc.stout.nau.util.Traversable.TraverseCallBack
	 */
	public boolean traverse(TraverseCallBack<NominalTerm> callBack) {
		if (!callBack.exec(this))
			if (propagate(callBack))
				return true;
		return callBack.execBackward(this);
	}

	/**
	 * Propagate traversing to sub-terms. Please override if needed!!!
	 * 
	 * @see Traversable
	 * @see at.jku.risc.stout.nau.util.Traversable.TraverseCallBack
	 */
	protected boolean propagate(TraverseCallBack<NominalTerm> callBack) {
		return false;
	}

}
