/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.nau.data.atom;

import java.io.IOException;
import java.io.Writer;
import java.util.Set;

import at.jku.risc.stout.nau.data.FreshnessCtx;

/**
 * A suspension is a {@linkplain NominalTerm} which consists of a
 * {@linkplain Permutation} and a {@linkplain Variable}.
 * 
 * @author Alexander Baumgartner
 */
public class Suspension extends NominalTerm {
	private Permutation perm;
	private Variable var;

	/**
	 * Instantiates a new suspension with an empty {@link Permutation}.
	 */
	public Suspension(Variable var) {
		this(var, new Permutation());
	}

	/**
	 * Instantiates a new suspension with the given {@link Permutation}. The
	 * given {@link Permutation} must not be null.
	 */
	public Suspension(Variable var, Permutation perm) {
		this.perm = perm;
		this.var = var;
	}

	@Override
	public boolean equals(Object other) {
		if (other instanceof Suspension) {
			Suspension oSusp = (Suspension) other;
			// Variable instances are unique
			return oSusp.var == var && oSusp.perm.equals(perm);
		}
		return false;
	}

	public Permutation getPerm() {
		return perm;
	}

	public void setPerm(Permutation perm) {
		this.perm = perm;
	}

	public Variable getVar() {
		return var;
	}

	public void setVar(Variable var) {
		this.var = var;
	}

	@Override
	public Sort getSort() {
		return var.getSort();
	}

	@Override
	public void setSort(Sort sort) throws MalformedSortException {
		var.setSort(sort);
	}

	@Override
	public void printString(Writer toPrint) throws IOException {
		if (!perm.isEmpty())
			toPrint.write(perm.toString());
		toPrint.write(var.toString());
	}

	@Override
	public NominalTerm swap(Atom a1, Atom a2) throws MalformedSortException {
		perm.addSwappingHead(a1, a2);
		return this;
	}

	@Override
	public boolean isFresh(Atom atom, FreshnessCtx nabla) {
		return nabla.contains(perm.permuteInverse(atom), var);
	}

	@Override
	public Suspension deepCopy() {
		return new Suspension(var, perm.deepCopy());
	}

	@Override
	public NominalTerm substitute(Variable fromVar, NominalTerm toTerm)
			throws MalformedSortException {
		if (var != fromVar)
			return this;
		if (toTerm.getSort() != getSort()) {
			if (toTerm.getSort() == null)
				toTerm.setSort(getSort());
			else if (getSort() != null)
				throw new MalformedSortException(
						"Cannot substitute for variable of different sort");
		}
		return toTerm.deepCopy().permute(perm);
	}

	@Override
	public NominalTerm permute(Permutation perm) {
		if (!perm.isEmpty())
			setPerm(Permutation.compose(perm, getPerm()));
		return this;
	}

	@Override
	public HasSort<Sort> getHead() {
		return var;
	}

	@Override
	public void collectAtoms(Set<Atom> atoms) {
		atoms.addAll(perm.getPerm().keySet());
	}
}
